/*
 * $Id: dcf77.c,v 1.4 2008/05/08 09:51:28 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18cxxx.h>
#include "NixieClock.h"
#include "dcf77.h"
#include "time.h"

/* DCF variables */
volatile time_t dcf_time;
volatile date_t dcf_date;
volatile unsigned char dcf_extra;
volatile dcf_flags_t dcf_flags;
volatile signed char dcf_bitcount;
volatile unsigned int dcf_high;
volatile unsigned int dcf_low;


static rom unsigned char dcf_bit_values[] =
	{ 1, 2, 4, 8, 10, 20, 40, 80 };

/* prototypes */
static void decode_bit(unsigned char);


void dcf77_enable(char enable)
{
	if (enable) {
		dcf_bitcount = 0;
		dcf_flags.value = 0;
		flags.do_dcf = 1;
	} else {
		flags.do_dcf = 0;
	}
}



/*
 * dcf77 interrupt is called every 100us
 */
void dcf77_fsm(void)
{
	/* display signal on LED */
	if (dcf_flags.synced) {
		LED_RED_PIN = 1;
		LED_GREEN_PIN = !DCF_PIN;
	} else {
		LED_RED_PIN = !DCF_PIN;
		LED_GREEN_PIN = 1;
	}

	/* decode signal */
	if (DCF_PIN) {
		/* pulse high */
		dcf_high++;
		if (dcf_low > DCF_2_S) {
			dcf_bitcount = 0;
			if (dcf_flags.full) {
				/* got a full frame */
				set_time(&dcf_time);
				set_date(&dcf_date);
				flags.do_dcf = 0;
				/* go back to "normal" mode */
				set_mode(MODE_AUTO);
			}
			/* reset values for next frame */
			dcf_time.hour = 0;
			dcf_time.min = 0;
			dcf_time.sec = 0;
			dcf_date.year = 0;
			dcf_date.month = 0;
			dcf_date.day = 0;
			dcf_date.weekday = 0;
			dcf_extra = 0;
			dcf_flags.value = 0;
			dcf_flags.synced = 1;
		}
		dcf_low = 0;
	} else {
		/* pulse low */
		dcf_low++;
		if (dcf_high > 0) {
			/* a high pulse was present */
			if (dcf_high > DCF_0_2_S) {
				/* we got a '1' */
				dcf_flags.parity ^= 1;
				decode_bit(1);
				dcf_bitcount++;
			} else if (dcf_high > DCF_0_1_S) {
				/* we got a '0' */
				decode_bit(0);
				dcf_bitcount++;
			} else {
				/* pulse too short, garbage */
				dcf_bitcount = 0;
				dcf_flags.synced = 0;
			}

			if (dcf_bitcount == 59) {
				/* received everything */
				dcf_flags.full = 1;
			}
		}
		dcf_high = 0;
	}
}


/*
 * Mark number(s) Encodes (01.s=0, 0.2s=1)
 *
 * 0              Minute, always 0 (0.1s)
 * 1-14           Reserved
 * 15             0=Normal antenna, 1=backup antenna
 * 16             1=Approaching change from MEZ to MESZ or back
 * 17,18          Time zone 0,1=MEZ; 1,0=MESZ
 * 19             The leap second is encoded in this bit one hour prior to
 *                occurrence.
 * 20             Start bit for encoded time, always 1
 * 21-27          1, 2, 4, 8, 10, 20, 40 Minutes (mark 21=1 minute)
 * 28             P1 maintains even parity for marks 21-28
 * 29-34          1,2,4,8,10,20 Hours (mark 29=1 hour)
 * 35             P2 maintains even parity for marks 29-35
 * 36-41          Day in month (1, 2, 4, 8, 10, 20)
 * 42-44          Day in week (1,2,4)
 * 45-49          Month number (1, 2, 4, 8, 10)
 * 50-57          Year (1, 2, 4, 8, 10, 20, 40, 80)
 * 58             P3 maintains even parity for marks 36-58 There is no mark
 *                transmitted for the 59th second.
 */
static void decode_bit(unsigned char b)
{
	signed char x;

	/* parity bits : 28, 35 and 58 */
	if (dcf_bitcount == 58 ||
	    dcf_bitcount == 35 ||
	    dcf_bitcount == 28) {
		/* parity should be 0 */
		if (dcf_flags.parity) {
			/* bad parity => reset */
			dcf_bitcount = 0;
			dcf_flags.synced = 0;
			return;
		}
		return;
	}

	if (b == 0) {
		return;
	}

	if (dcf_bitcount >= 50) {
		dcf_date.year += dcf_bit_values[dcf_bitcount - 50];
		return;
	}
	if (dcf_bitcount >= 45) {
		dcf_date.month += dcf_bit_values[dcf_bitcount - 45];
		return;
	}
	if (dcf_bitcount >= 42) {
		dcf_date.weekday += dcf_bit_values[dcf_bitcount - 42];
		return;
	}
	if (dcf_bitcount >= 36) {
		dcf_date.day += dcf_bit_values[dcf_bitcount - 36];
		return;
	}
	if (dcf_bitcount >= 29) {
		dcf_time.hour += dcf_bit_values[dcf_bitcount - 29];
		return;
	}
	if (dcf_bitcount >= 21) {
		dcf_time.min += dcf_bit_values[dcf_bitcount - 21];
		return;
	}
	if (dcf_bitcount >= 15) {
		dcf_extra |= (1 << (dcf_bitcount - 15));
	}
	dcf_flags.parity = 0;

	return;
}

