/*
 * $Id: ds1307.c,v 1.6 2008/05/08 09:51:28 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18cxxx.h>
#include "NixieClock.h"
#include "sw_i2c.h"
#include "ds1307.h"
#include "time.h"
#include "command.h"
#include "util.h"


int ds1307_read(unsigned char address)
{
	unsigned int value;

	i2c_start();
	if (i2c_putc(DS1307_SLA)) {
		return (-1);
	}
	if (i2c_putc(address)) {
		return (-1);
	}
	i2c_restart();
	if (i2c_putc(DS1307_SLA | 0x1)) {
		return (-1);
	}
	value = i2c_getc();
	i2c_stop();

	return (value);
}


signed char ds1307_write(unsigned char address, unsigned char data)
{
	i2c_start();
	if (i2c_putc(DS1307_SLA)) {
		return (-1);
	}
	if (i2c_putc(address)) {
		return (-1);
	}
	if (i2c_putc(data)) {
		return (-1);
	}
	i2c_stop();

	return (0);
}


/*
 * Initialize DS1307 :
 * - set CH to 0    (enable oscillator)
 * - set SQWE to 1  (enable clock output)
 * - set oscillator output to 32768Hz
 * - set hour to 24h mode
 */
signed char ds1307_init(void)
{
	unsigned int x;
	signed char ret;

	x = ds1307_read(DS1307_MAGIC_LOC);
	if (x & 0xFF00) {
		return (-1);
	}

	if (x == DS1307_MAGIC) {
		/* already initialized */
		return (0);
	}

	/* beep to signal initialize */
	DOUBLE_BEEP;

	ret = ds1307_write(DS1307_SEC, 0);
	ret |= ds1307_write(DS1307_MIN, 0);
	ret |= ds1307_write(DS1307_HOUR, 0);
	ret |= ds1307_write(DS1307_WEEKDAY, 1);
	ret |= ds1307_write(DS1307_DAY, 1);
	ret |= ds1307_write(DS1307_MONTH, 1);
	ret |= ds1307_write(DS1307_YEAR, 8);
	ret |= ds1307_write(DS1307_CONTROL, DS1307_CONFIG);
	ret |= ds1307_write(DS1307_MAGIC_LOC, DS1307_MAGIC);

	/* initialize tube_time to 0day 0hour 0min */
	tube_time.day = 0;
	tube_time.hour = 0;
	tube_time.min = 0;
	tube_time.sec = 0;
	ds1307_save_tube_time();

	return (ret);
}


/*
 * Read time from DS1307
 */
signed char ds1307_get_time(volatile time_t *x)
{
	unsigned int ret;

	ret = ds1307_read(DS1307_SEC);
	if (ret < 0) {
		return (-1);
	}
	x->sec = unbcd((unsigned char)ret);

	ret = ds1307_read(DS1307_MIN);
	if (ret < 0) {
		return (-1);
	}
	x->min = unbcd((unsigned char)ret);

	ret = ds1307_read(DS1307_HOUR);
	if (ret < 0) {
		return (-1);
	}
	x->hour = unbcd((unsigned char)ret);

	return (0);
}

/*
 * Set time in DS1307
 */
signed char ds1307_set_time(volatile time_t *x)
{
	unsigned char ret;

	ret = ds1307_write(DS1307_SEC, bcd(x->sec));
	ret |= ds1307_write(DS1307_MIN, bcd(x->min));
	ret |= ds1307_write(DS1307_HOUR, bcd(x->hour));

	return (ret);
}


/*
 * Read date from DS1307
 */
signed char ds1307_get_date(volatile date_t *x)
{
	unsigned int ret;

	ret = ds1307_read(DS1307_YEAR);
	if (ret < 0) {
		return (-1);
	}
	x->year = unbcd((unsigned char)ret);

	ret = ds1307_read(DS1307_MONTH);
	if (ret < 0) {
		return (-1);
	}
	x->month = unbcd((unsigned char)ret);

	ret = ds1307_read(DS1307_DAY);
	if (ret < 0) {
		return (-1);
	}
	x->day = unbcd((unsigned char)ret);

	ret = ds1307_read(DS1307_WEEKDAY);
	if (ret < 0) {
		return (-1);
	}
	x->weekday = unbcd((unsigned char)ret);

	return (0);
}


/*
 * Set date in DS1307
 */
signed char ds1307_set_date(volatile date_t *x)
{
	signed char ret;

	ret = ds1307_write(DS1307_YEAR, bcd(x->year));
	ret |= ds1307_write(DS1307_MONTH, bcd(x->month));
	ret |= ds1307_write(DS1307_DAY, bcd(x->day));
	ret |= ds1307_write(DS1307_WEEKDAY, bcd(x->weekday));

	return (ret);
}


/*
 * Get the tube time from DS1307
 *
 * Address -  Data
 * ================
 * 0x0A		day (upper byte)
 * 0x0B		day (lower byte)
 * 0x0C		hour
 * 0x0D		min
 */
signed char ds1307_restore_tube_time(void)
{
	unsigned int ret;

	ret = ds1307_read(0xA);
	if (ret < 0) {
		return (-1);
	}
	tube_time.day = ret << 8;

	ret = ds1307_read(0xB);
	if (ret < 0) {
		return (-1);
	}
	tube_time.day += ret;

	ret = ds1307_read(0x0C);
	if (ret < 0) {
		return (-1);
	}
	tube_time.hour = ret;

	ret = ds1307_read(0x0D);
	if (ret < 0) {
		return (-1);
	}

	tube_time.sec = 0;

	return (0);
}


/*
 * Save tube time in DS1307
 */
signed char ds1307_save_tube_time(void)
{
	signed char ret;

	ret = ds1307_write(0xA,
		(unsigned char)(tube_time.day >> 8));
	ret |= ds1307_write(0xB,
		(unsigned char)(tube_time.day & 0xFF));
	ret |= ds1307_write(0xC, tube_time.hour);
	ret |= ds1307_write(0xD, tube_time.min);

	return (ret);
}

