/*
 * $Id: editor.c,v 1.24 2008/06/11 15:44:07 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18cxxx.h>
#include "NixieClock.h"
#include "editor.h"
#include "tubes.h"
#include "time.h"
#include "util.h"
#include "command.h"
#include "version.h"


/* prototypes */
static unsigned int editor2int(void);

static char int_valid(void);
static char time4_valid(void);
static char time6_valid(void);
static char date6_valid(void);

static void ed_nothing(void);
static void ed_get_time(void);
static void ed_set_time(void);
static void ed_get_date(void);
static void ed_set_date(void);
static void ed_get_tube_time(void);
static void ed_get_alarm(void);
static void ed_set_alarm(void);
static void ed_get_alarm_enable(void);
static void ed_set_alarm_enable(void);
static void ed_get_alarm_duration(void);
static void ed_set_alarm_duration(void);
static void ed_get_brightness(void);
static void ed_set_brightness(void);
static void ed_get_tube_auto(void);
static void ed_set_tube_auto(void);
static void ed_get_tube_off(void);
static void ed_set_tube_off(void);
static void ed_get_tube_on(void);
static void ed_set_tube_on(void);
static void ed_get_date_format(void);
static void ed_set_date_format(void);
static void ed_get_date_when(void);
static void ed_set_date_when(void);
static void ed_get_date_duration(void);
static void ed_set_date_duration(void);
static void ed_get_hour_zero(void);
static void ed_set_hour_zero(void);
static void ed_get_prevent_time(void);
static void ed_set_prevent_time(void);
static void ed_get_prevent_duration(void);
static void ed_set_prevent_duration(void);
static void ed_get_version(void);


static rom  menu_t menus[] = {

	// 0  alarm on/off   0 = off   1 = on
	{ TYPE_1_DIGIT, 0, 1, 1, ed_get_alarm_enable, ed_set_alarm_enable, int_valid },

	// 1  alarm time   HH:MM
	{ TYPE_4_TIME, 0, 0, 1, ed_get_alarm, ed_set_alarm, time4_valid },

	// 2  alarm duration  1-30 sec
	{ TYPE_2_DIGIT, 1, 30, 1, ed_get_alarm_duration, ed_set_alarm_duration, int_valid },

	// 3  default brightness  0-7, 8 = auto
	{ TYPE_1_DIGIT, 0, 8, 1, ed_get_brightness, ed_set_brightness , int_valid },

	// 4  tube auto on/off    0 = always on  1 = auto off, no DCF   2 = auto off, DCF
	{ TYPE_1_DIGIT, 0, 2, 1, ed_get_tube_auto, ed_set_tube_auto, int_valid },

	// 5  tube off time  HH:MM
	{ TYPE_4_TIME, 0, 0, 1, ed_get_tube_off, ed_set_tube_off, time4_valid },

	// 6  tube on time   HH:MM
	{ TYPE_4_TIME, 0, 0, 1,  ed_get_tube_on, ed_set_tube_on, time4_valid },

	// 7 cathode poisoning duration  0 = don't do it  1-99 duration per digit in 10sec
	{ TYPE_2_DIGIT, 0, 99, 1, ed_get_prevent_duration, ed_set_prevent_duration, int_valid },

	// 8 cathode poisoning time
	{ TYPE_4_TIME, 0, 0, 1,  ed_get_prevent_time, ed_set_prevent_time, time4_valid },

	// 9  date format 0 = DD.MM.YY   1 = MM.DD.YY
	{ TYPE_1_DIGIT, 0, 1, 1, ed_get_date_format, ed_set_date_format, int_valid },

	// 10  show date duration    0 = don't show   1-58 sec
	{ TYPE_2_DIGIT, 0, 58, 1, ed_get_date_duration, ed_set_date_duration, int_valid },

	// 11 show date when   0-59 second when date is first shown
	{ TYPE_2_DIGIT, 0, 59, 1, ed_get_date_when, ed_set_date_when, int_valid },

	// 12 hour leading zero suppression  0 = don't suppress   1 = suppress
	{ TYPE_1_DIGIT, 0, 1, 1, ed_get_hour_zero, ed_set_hour_zero, int_valid },

	// 13  firmware version (not editable)
	{ TYPE_6_DIGIT, 0, 0, 0, ed_get_version, ed_nothing, ed_nothing },

	// 14  tube total on time (not editable)
	{ TYPE_6_DIGIT, 0, 0, 0, ed_get_tube_time, ed_nothing, ed_nothing },

	// 15  current time  HH:MM:SS
	{ TYPE_6_TIME, 0, 0, 1, ed_get_time, ed_set_time, time6_valid },

	// 16  current date  DD.MM.YY  (always edited in this format)
	{ TYPE_6_DATE, 0, 0, 1, ed_get_date, ed_set_date, date6_valid }

};

#define	NUM_MENUS	17



/* variables */
signed char ed_current;		// current menu
unsigned char ed_timeout;	// time (in sec) before editor is
				// exited.
signed char ed_tube_cur;	// tube being edited (blink)
signed char ed_first;		// first tube editable
signed char ed_x[6];		// value being edited.


/* init editor related variables */
void editor_init(void)
{
	ed_current = 0;
	ed_tube_cur = -1;
	editor_get();
}

/* called when the editor timeout has elapsed */
void editor_timeout(void)
{
	ed_tube_cur = -1;
	set_mode(MODE_TIME);
}

/* get the data associated to a menu, and update the
 * tube values
 */
void editor_get(void)
{
	static char i;

#if USE_DP
	/* reset decimal points */
	tubes_dp = 0;
#endif

	/* get current value */
	(*(menus[ed_current].f_get))();

	/* compute first editable tube */
	i = menus[ed_current].type;
	if (i <= TYPE_6_DIGIT) {
		ed_first = 5 - i;
	} else if (i == TYPE_4_TIME) {
		ed_first = 2;
	} else {
		ed_first = 0;
	}

	/*
	 * tubes after menu number and before first data tube
	 * are off
	 */
	for (i = 2; i < ed_first; i++) {
		ed_x[i] = 0xFF;
	}

	ed_tube_cur = -1;
}


/* set the tubes to the correct values */
void editor_refresh(void)
{
	static char i;

	/* display menu number on the left-most tube pair */
	if (ed_first >= 2) {
		if (ed_current > 9) {
			tubes_set_pair(T_BOTH, 0, ed_current, 0);
#if USE_DP
			if (menus[ed_current].editable) {
				tubes_dp |= 0x2;
			}
#endif
		} else {
			tubes_set_one(T_BOTH, 1, 0xFF);
			tubes_set_one(T_BOTH, 0, ed_current);
#if USE_DP
			if (menus[ed_current].editable) {
				tubes_dp |= 0x1;
			}
#endif
		}
	}

	/* display current edited value */
	for (i = (ed_first >= 2 ? 2 : 0); i < 6; i++) {
		tubes_set_one(T_BOTH, i, ed_x[i]);
	}

	/* blinking tube */
	if (ed_tube_cur >= 0) {
		tubes_set_one(T_2, ed_tube_cur, 0xFF);
	}

	ed_timeout = EDITOR_TIMEOUT;
}


void editor_adjust_menu(signed char x)
{
	ed_current = (ed_current + x) % NUM_MENUS;
	if (ed_current < 0) {
		ed_current += NUM_MENUS;
	}
	ed_tube_cur = -1;
	editor_get();
	editor_refresh();
}


/* ENTER pressed while in editor mode */
void editor_enter(void)
{
	if (!menus[ed_current].editable) {
		/* value is not editable */
		SINGLE_BEEP;
		return;
	}

	if (ed_tube_cur < 0) {
		/* nothing under edit */
		ed_tube_cur = ed_first;
	} else {
		/* user has accepted the current value */
		(*(menus[ed_current].f_set))();
		DOUBLE_BEEP;
		ed_tube_cur = -1;
	}
	editor_refresh();
}


/* UP or DOWN pressed */
void editor_up_down(signed char x)
{
	if (ed_tube_cur < 0) {
		/* nothing under edit */
		if (x > 0) {
			editor_adjust_menu(1);
		} else {
			editor_adjust_menu(-1);
		}
	} else {
		/* we are editing. go one char left or right */
		if (x > 0) {
			editor_adjust_digit(1, 1);
		} else {
			editor_adjust_digit(-1, 1);
		}
	}
}


/* number 0-9 has been pressed */
void editor_num(char key, char auto_advance)
{
	static char prev;

	if (ed_tube_cur < 0) {
		/* nothing under edit => invalid command */
		SINGLE_BEEP;
		return;
	}

	prev = ed_x[ed_tube_cur];
	ed_x[ed_tube_cur] = key;
	if ((*(menus[ed_current].f_valid))()) {
		if (auto_advance) {
			editor_adjust_digit(1, 0);
		}
		editor_refresh();
	} else {
		/* not valid */
		ed_x[ed_tube_cur] = prev;
		SINGLE_BEEP;
	}
}


/* increment or decrement current tube under edit */
void editor_inc_digit(signed char x)
{
	static char key;

	key = ed_x[ed_tube_cur];
	if (x > 0 && key < 9) {
		editor_num(key + 1, 0);
		return;
	}
	if (x < 0 && key > 0) {
		editor_num(key - 1, 0);
		return;
	}
	SINGLE_BEEP;
}


/* move edited tube +1 or -1. */
void editor_adjust_digit(signed char x, char beep_on_error)
{
	if (ed_tube_cur != -1) {
		if (x > 0) {
			if (ed_tube_cur < 5) {
				ed_tube_cur++;
				editor_refresh();
				return;
			}
		} else {
			if (ed_tube_cur > ed_first) {
				ed_tube_cur--;
				editor_refresh();
				return;
			}
		}
	}

	if (beep_on_error) {
		SINGLE_BEEP;
	}
}


/* get the integer currently edited (values 0000 -> 9999) */
static unsigned int editor2int(void)
{
	unsigned int x;
	static const unsigned int v[] = { 1000, 100, 10, 1 };
	char i;

	x = 0;
	for (i = ed_first; i < 6; i++) {
		x += ed_x[i] * v[i - 2];
	}
	return (x);
}


/* check if an integer is valid for the current menu */
static char int_valid(void)
{
	unsigned int x;

	x = editor2int();
	if (x < menus[ed_current].x_min) {
		return (0);
	}
	if (x > menus[ed_current].x_max) {
		return (0);
	}

	return (1);
}


/* check if the current value is between 00:00 and 23:59 */
static char time4_valid(void)
{
	if ((ed_x[2] * 10 + ed_x[3]) > 23) {
		/*
		 * be smart for the user : if the current tube being
		 * edited is the tens from the hours, and if it's valid
		 * (0-2), then automatically set a valid number for
		 * the ones of the hours.
		 *
		 * Example: current value is 19:54, and "1" blinks.
		 *          the user cannot enter "2" because 29:54 is
		 *          invalid. So if "2" is typed, we set 20:54
		 *          automatically.
		 */
		if (ed_tube_cur == 2 && ed_x[2] < 3) {
			ed_x[3] = 0;
			return (1);
		}
		return (0);
	}
	if ((ed_x[4] * 10 + ed_x[5]) > 59) {
		return (0);
	}
	return (1);
}


/* check if the current value is between 00:00:00 and 23:59:59 */
static char time6_valid(void)
{
	if ((ed_x[0] * 10 + ed_x[1]) > 23) {
		if (ed_tube_cur == 0 && ed_x[0] < 3) {
			ed_x[1] = 0;
			return (1);
		}
		return (0);
	}
	if ((ed_x[2] * 10 + ed_x[3]) > 59) {
		return (0);
	}
	if ((ed_x[4] * 10 + ed_x[5]) > 59) {
		return (0);
	}
	return (1);
}


/* check if the current value is a valid date DD.MM.YY */
static char date6_valid(void)
{
	char day;
	char month;
	char x;
	char leap;
	static const char months[] =
		{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

	day = ed_x[0] * 10 + ed_x[1];
	month = ed_x[2] * 10 + ed_x[3];
	x = ed_x[4] * 10 + ed_x[5];		// x : year
	leap = (x % 4 == 0);

	if (ed_tube_cur == 0) {
		/* user is editing the tens of the days */
		if (ed_x[0] == 0 && ed_x[1] == 0) {
			ed_x[1] = 1;	// 00 -> 01
		} else if (ed_x[0] == 3 && ed_x[1] > 1) {
				ed_x[1] = 0; // 3[2-9] -> 30
		}
		day = ed_x[0] * 10 + ed_x[1];
	}
	if (day == 0 || day > 31) {
		return (0);
	}

	/* auto-correct month */
	if (ed_tube_cur == 2) {
		if (ed_x[2] == 0 && ed_x[3] == 0) {
			ed_x[3] = 1;	// 00 -> 01
		} else if (ed_x[2] == 1 && ed_x[3] > 2) {
			ed_x[3] = 0;	// 1[3-9] -> 10
		}
		month = ed_x[2] * 10 + ed_x[3];
	}
	if (month == 0 || month > 12) {
		return (0);
	}

	/* check day/month is valid */
	x = months[month - 1];
	if (month == 2 && leap) x++;

	if (day > x) {
		if (ed_tube_cur >= 3) {
			/* adapt 31.04.08 -> 30.04.08 automatically */
			day = bcd(x);
			ed_x[0] = day >> 4;
			ed_x[1] = day & 0x0F;
		} else if (ed_tube_cur == 2) {
			/* 31.11.08 -> 31.10.08 (editing tens of month) */
			ed_x[3] = 0;
		} else {
			/* user is editing day. change month instead */
			ed_x[2] = 0;
			ed_x[3] = 1;
		}
	}

	return (1);
}


/* utility function to set the value of 2 tubes (editor mode) */
static void set_bcd(signed char *p, unsigned char value)
{
	value = bcd(value);
	p[0] = value >> 4;
	p[1] = value & 0x0F;
}


/* empty fonction */
static void ed_nothing(void)
{
}


/* get current time and copy it to the editor */
static void ed_get_time(void)
{
	set_bcd(ed_x, time.hour);
	set_bcd(ed_x + 2, time.min);
	set_bcd(ed_x + 4, time.sec);
}


/* set the current time from the editor */
static void ed_set_time(void)
{
	volatile time_t t;

	t.hour = ed_x[0] * 10 + ed_x[1];
	t.min = ed_x[2] * 10 + ed_x[3];
	t.sec = ed_x[4] * 10 + ed_x[5];

	set_time(&t);
}


/* get current date and copy it to the editor */
static void ed_get_date(void)
{
	set_bcd(ed_x, date.day);
	set_bcd(ed_x + 2, date.month);
	set_bcd(ed_x + 4, date.year);
}


/* set the current date from the editor */
static void ed_set_date(void)
{
	volatile date_t d;

	d.day = ed_x[0] * 10 + ed_x[1];
	d.month = ed_x[2] * 10 + ed_x[3];
	d.year = ed_x[4] * 10 + ed_x[5];

	set_date(&d);
}


/* get alarm time */
static void ed_get_alarm(void)
{
	set_bcd(ed_x + 2, alarm.hour);
	set_bcd(ed_x + 4, alarm.min);
}


/* set alarm time */
static void ed_set_alarm(void)
{
	alarm.hour = ed_x[2] * 10 + ed_x[3];
	alarm.min = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_ALARM_HH, alarm.hour);
	ee_write(ADR_ALARM_MM, alarm.min);

	/* alarm is set. enable it automatically */
	prefs.alarm_enable = 1;
	ee_write(ADR_PREFS, prefs.value);
}


/* get default brightness */
static void ed_get_brightness(void)
{
	ed_x[5] = ee_read(ADR_BRIGHTNESS);
}


/* set default brightness */
static void ed_set_brightness(void)
{
	ee_write(ADR_BRIGHTNESS, ed_x[5]);
	tubes_brightness(ed_x[5]);
}


/* get alarm enable */
static void ed_get_alarm_enable(void)
{
	ed_x[5] = prefs.alarm_enable ? 1 : 0;
}


/* set alarm enable */
static void ed_set_alarm_enable(void)
{
	prefs.alarm_enable = ed_x[5] ? 1 : 0;

	ee_write(ADR_PREFS, prefs.value);
}


/* get alarm duration */
static void ed_get_alarm_duration(void)
{
	set_bcd(ed_x + 4, alarm_duration);
}

/* set alarm duration */
static void ed_set_alarm_duration(void)
{
	unsigned char x;

	alarm_duration = (unsigned char)editor2int();

	ee_write(ADR_ALARM_DURATION, alarm_duration);
}

/* get tube save */
static void ed_get_tube_auto(void)
{
	if (prefs.tube_save) {
		if (prefs.tube_save_dcf) {
			ed_x[5] = 2;
		} else {
			ed_x[5] = 1;
		}
	} else {
		ed_x[5] = 0;
	}
}

/* set tube save */
static void ed_set_tube_auto(void)
{
	switch (ed_x[5]) {

	case 0:
		prefs.tube_save = 0;
		break;
	case 1:
		prefs.tube_save = 1;
		prefs.tube_save_dcf = 0;
		break;
	case 2:
		prefs.tube_save = 1;
		prefs.tube_save_dcf = 1;
	}

	ee_write(ADR_PREFS, prefs.value);
}


/* get tube off time */
static void ed_get_tube_off(void)
{
	set_bcd(ed_x + 2, tube_off.hour);
	set_bcd(ed_x + 4, tube_off.min);
}


/* set tube off time */
static void ed_set_tube_off(void)
{
	tube_off.hour = ed_x[2] * 10 + ed_x[3];
	tube_off.min = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_TUBE_OFF_HH, tube_off.hour);
	ee_write(ADR_TUBE_OFF_MM, tube_off.min);
}

/* get tube on time */
static void ed_get_tube_on(void)
{
	set_bcd(ed_x + 2, tube_on.hour);
	set_bcd(ed_x + 4, tube_on.min);
}

/* set tube on time */
static void ed_set_tube_on(void)
{
	tube_on.hour = ed_x[2] * 10 + ed_x[3];
	tube_on.min = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_TUBE_ON_HH, tube_on.hour);
	ee_write(ADR_TUBE_ON_MM, tube_on.min);
}

/* get date format */
static void ed_get_date_format(void)
{
	ed_x[5] = prefs.date_format ? 1 : 0;
}

/* set date format */
static void ed_set_date_format(void)
{
	prefs.date_format = ed_x[5] ? 1 : 0;

	ee_write(ADR_PREFS, prefs.value);
}

/* get date duration */
static void ed_get_date_duration(void)
{
	set_bcd(ed_x + 4, show_date_duration);
}


/* set date duration */
static void ed_set_date_duration(void)
{
	show_date_duration = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_DATE_DURATION, show_date_duration);
}


/* get date when */
static void ed_get_date_when(void)
{
	set_bcd(ed_x + 4, show_date_when);
}


/* set date when */
static void ed_set_date_when(void)
{
	show_date_when = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_DATE_WHEN, show_date_when);
}


/* get hour leading zero suppress */
static void ed_get_hour_zero(void)
{
	ed_x[5] = prefs.hour_leading_zero ? 1 : 0;
}

/* set hour leading zero suppress */
static void ed_set_hour_zero(void)
{
	prefs.hour_leading_zero = ed_x[5] ? 1 : 0;

	ee_write(ADR_PREFS, prefs.value);
}



/* get prevention tube */
static void ed_get_prevent_time(void)
{
	set_bcd(ed_x + 2, t_when.hour);
	set_bcd(ed_x + 4, t_when.min);
}


/* set prevention time */
static void ed_set_prevent_time(void)
{
	t_when.hour = ed_x[2] * 10 + ed_x[3];
	t_when.min = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_PREVENT_HH, t_when.hour);
	ee_write(ADR_PREVENT_MM, t_when.min);
}


/* get prevent duration */
static void ed_get_prevent_duration(void)
{
	set_bcd(ed_x + 4, t_time_reload);
}


/* set date when */
static void ed_set_prevent_duration(void)
{
	t_time_reload = ed_x[4] * 10 + ed_x[5];

	ee_write(ADR_PREVENT_RELOAD, t_time_reload);
}


/* get firmware version */
static void ed_get_version(void)
{
	ed_x[0] = 0xff;
	ed_x[1] = 0xff;
	set_bcd(ed_x + 2, VERSION_MAIN);
	if (ed_x[2] == 0) ed_x[2] = 0xff;
	set_bcd(ed_x + 4, VERSION_MINOR);
#if USE_DP
	tubes_dp |= 0x10;
#endif
}


/* get tube time DDDD.HH  where DDDD=number of days, HH = number of hours */
static void ed_get_tube_time(void)
{
	static unsigned int x;

	x = tube_time.day;

	ed_x[0] = 0;
	ed_x[1] = 0;
	ed_x[2] = 0;
	ed_x[3] = 0;
	while (x >= 1000) {
		x -= 1000;
		ed_x[0]++;
	}
	while (x >= 100) {
		x -= 100;
		ed_x[1]++;
	}
	while (x >= 10) {
		x -= 10;
		ed_x[2]++;
	}
	ed_x[3] = x;

	/* remove leading zero */
	if (ed_x[0] == 0) {
		ed_x[0] = 0xff;
		if (ed_x[1] == 0) {
			ed_x[1] = 0xff;
			if (ed_x[2] == 0) {
				ed_x[2] = 0xff;
			}
		}
	}

	set_bcd(ed_x + 4, tube_time.hour);

#if USE_DP
	tubes_dp |= 0x10;
#endif
}



/* copy an integer to the editor (values 0000 -> 9999) */
static void int2editor(unsigned int x)
{

}

