/*
 * $Id: main.c,v 1.23 2008/05/11 08:45:41 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18f2431.h>
#include <timers.h>
#include <delays.h>
#include "NixieClock.h"
#include "ds1307.h"
#include "dcf77.h"
#include "nec.h"
#include "util.h"
#include "tubes.h"
#include "sw_i2c.h"
#include "command.h"
#include "editor.h"

/* microcontroller config */
#pragma config OSC=HSPLL
#pragma config PWRTEN=ON, BOREN=ON, MCLRE=ON, LVP=OFF
#pragma config WDTEN=OFF, STVREN=ON

/* globals */
volatile flags_t flags;
volatile prefs_t prefs;
volatile unsigned char alarm_duration;
volatile unsigned char show_date_when;
volatile unsigned char show_date_duration;
volatile unsigned char mode;
volatile int temp_show;


/* prototypes */
void handle_interrupt(void);
static void handle_timer0(void);
static void handle_timer1(void);



/* interrupt vector */
#pragma code interrupt_code = 0x08
void interrupt(void)
{
	_asm
		goto handle_interrupt
	_endasm
}


/* interrupt handler */
#pragma code
#pragma interrupt handle_interrupt
void handle_interrupt(void)
{
	if (PIR1bits.TMR1IF) {
		handle_timer1();
		PIR1bits.TMR1IF = 0;
	} else if (INTCONbits.INT0IF) {
		handle_nec_interrupt();
		INTCONbits.INT0IF = 0;
	} else if (INTCONbits.TMR0IF) {
		handle_timer0();
		INTCONbits.TMR0IF = 0;
	}
}


/*
 * Timer0 overflows every 100uS
 */
static void handle_timer0(void)
{
	/*
	 * IR decoding requires a counter to measure
	 * pulses length
	 */
	if (IR_count == 255) {
		/* timeout occured, reset IR decode routine */
		nec_reset();
	} else {
		IR_count++;
	}

	/* DCF77 */
	if (flags.do_dcf) {
		dcf77_fsm();
	}

	/* refresh tubes */
	if (flags.do_tube) {
		tubes_refresh();
	}

	/* reload timer 0 */
	TMR0H = TMR0_H_RELOAD;
	TMR0L = TMR0_L_RELOAD;
}


/*
 * Timer1 overflows every 1/256s
 */
static void handle_timer1(void)
{
	static unsigned int prod;

	/* reload timer 1 */
	TMR1H = TMR1_H_RELOAD;
	TMR1L = TMR1_L_RELOAD;

	/* increment tick (1/256s) */
	tick++;
	if (tick == 0) {
		time_inc();
	}

	/* decrement temp_show */
	if (temp_show != 0) {
		temp_show--;
	}

	/* alarm & beep */
	if (flags.do_alarm) {
		if (beep_duration != 0) {
			if ((tick & 0xC0) == 0x40 ||
			    (tick & 0xC0) == 0xC0) {
				ALARM_PIN = 1;
			} else {
				ALARM_PIN = 0;
			}

			if (tick == 0) {
				beep_duration--;
			}
		} else {
			ALARM_PIN = 0;
			flags.do_alarm = 0;
		}
	} else {
		/* handle beeping */
		if (beep > 0) {
			/*
			 * - beep for 16 cycles at a time.
			 * - to generate a simple beep, beep = 16
			 * - to generate a double beep, beep = 48
			 */
			beep--;
			ALARM_PIN = (beep & 0x10) ? 0 : 1;
		} else {
			ALARM_PIN = 0;
		}
	}

	/* automatic brightness (every 1/4s) */
	if (prefs.auto_brightness && !(tick & 0x1F)) {
		flags.do_ad = 1;
	}

	/* update fading threshold */
	if (tick < 128) {
		fade_threshold = 0;
	} else {
		prod = tick & 0x7F;
		prod *= brightness_refresh;
		prod >>= 6;
		fade_threshold = (unsigned char)prod;
	}
}



static void read_prefs(void)
{
	if (ee_read(ADR_MAGIC) != PREFS_MAGIC) {

		prefs.value = 0;
		prefs.tube_save = 1;
		prefs.tube_save_dcf = 1;
		brightness = 5;
		show_date_when = 45;
		show_date_duration = 0;
		alarm.hour = 7;
		alarm.min = 0;
		alarm_duration = 9;
		tube_on.hour = 7;
		tube_on.min = 0;
		tube_off.hour = 22;
		tube_off.min = 0;
		t_time_reload = 6;	// in 10sec
		t_when.hour = 3;
		t_when.min = 0;

		ee_write(ADR_PREFS, prefs.value);
		ee_write(ADR_BRIGHTNESS, brightness);
		ee_write(ADR_DATE_WHEN, show_date_when);
		ee_write(ADR_DATE_DURATION, show_date_duration);
		ee_write(ADR_ALARM_HH, alarm.hour);
		ee_write(ADR_ALARM_MM, alarm.min);
		ee_write(ADR_ALARM_DURATION, alarm_duration);
		ee_write(ADR_TUBE_ON_HH, tube_on.hour);
		ee_write(ADR_TUBE_ON_MM, tube_on.min);
		ee_write(ADR_TUBE_OFF_HH, tube_off.hour);
		ee_write(ADR_TUBE_OFF_MM, tube_off.min);
		ee_write(ADR_PREVENT_RELOAD, t_time_reload);
		ee_write(ADR_PREVENT_HH, t_when.hour);
		ee_write(ADR_PREVENT_MM, t_when.min);

		ee_write(ADR_MAGIC, PREFS_MAGIC);
	}

	prefs.value = ee_read(ADR_PREFS);
	show_date_when = ee_read(ADR_DATE_WHEN);
	show_date_duration = ee_read(ADR_DATE_DURATION);
	brightness = ee_read(ADR_BRIGHTNESS);
	alarm.hour = ee_read(ADR_ALARM_HH);
	alarm.min = ee_read(ADR_ALARM_MM);
	alarm_duration = ee_read(ADR_ALARM_DURATION);
	tube_on.hour = ee_read(ADR_TUBE_ON_HH);
	tube_on.min = ee_read(ADR_TUBE_ON_MM);
	tube_off.hour = ee_read(ADR_TUBE_OFF_HH);
	tube_off.min = ee_read(ADR_TUBE_OFF_MM);
	t_time_reload = ee_read(ADR_PREVENT_RELOAD);
	t_when.hour = ee_read(ADR_PREVENT_HH);
	t_when.min = ee_read(ADR_PREVENT_MM);
}



static void configure_ports(void)
{
	/* configure port A */
	ALARM_TRIS = 0;
	ALARM_ANSEL = 0;
	ALARM_PIN = 0;

	LED_RED_TRIS = 0;
	LED_RED_ANSEL = 0;
	LED_RED_PIN = 1;

	LED_GREEN_TRIS = 0;
	LED_GREEN_ANSEL = 0;
	LED_GREEN_PIN = 1;

	DCF_TRIS = 1;
	DCF_ANSEL = 0;

	LDR_TRIS = 1;
	LDR_ANSEL = 1;
	ADCHS = 0x01;
	ADCON0bits.ACMOD0 = 0;
	ADCON0bits.ACMOD1 = 0;
	ADCON0bits.ADON = 1;
	ADCON2 = 0b00000110;


	/* configure port B */
	TRISB = 0;

	/* configure port C */
	TRISC = 0xFF;

	HV_TRIS = 0;
	HV_OFF;

	ANODE0_TRIS = 0;
	ANODE0 = 0;
	ANODE1_TRIS = 0;
	ANODE1 = 0;
	ANODE2_TRIS = 0;
	ANODE2 = 0;

	i2c_init();
}


void main(void)
{
	Delay100TCYx(10);

	/* configure ports */
	configure_ports();

	/* configure timers */
	OpenTimer0(TIMER_INT_ON & T0_16BIT & T0_SOURCE_INT &
		T0_PS_1_1);
	TMR0H = TMR0_H_RELOAD;
	TMR0L = TMR0_L_RELOAD;

	OpenTimer1(TIMER_INT_ON & T1_16BIT_RW & T1_SOURCE_EXT &
		T1_PS_1_1 & T1_SYNC_EXT_OFF & T1_OSC1EN_OFF);
	TMR1H = TMR1_H_RELOAD;
	TMR1L = TMR1_L_RELOAD;

	/* reset flags */
	flags.value = 0;

	/* read preferences */
	read_prefs();

	/* enable IR reception */
	nec_enable();

	/* initialize editor */
	editor_init();

	/* set brightness */
	tubes_brightness(brightness);

	/* initialize DS1307 */
	ds1307_init();
	ds1307_get_date(&date);
	ds1307_restore_tube_time();
	ds1307_get_time(&time);

	/*
	 * default mode : auto, but overwrite AUTO_OFF to
	 *  make sure we show something on the tubes when the
	 *  clock is turned on
	 */
	set_mode(MODE_AUTO);
	if (mode == MODE_AUTO_OFF) {
		set_mode(MODE_TIME);
	}


	/* enable interrupts */
	INTCONbits.PEIE = 1;  		// required for TMR1 interrupt
	INTCONbits.GIE = 1;

#if 0
	SINGLE_BEEP;
#endif

	/* main loop */
	while (1) {

		/* handle auto-brightness if needed */
		if (flags.do_ad) {
			flags.do_ad = 0;
			ADCON0bits.GO = 1;
			while (ADCON0bits.GO);
			tubes_brightness(ADRESH >> 6);
		}

		/* handle keyboard */
		if (flags.do_command) {
			flags.do_command = 0;
			handle_command();
		}

		/* save tube time in nvram */
		if (flags.do_save_tube_time) {
			flags.do_save_tube_time = 0;
			ds1307_save_tube_time();
		}

		/* minute work */
		if (flags.do_minute_work) {
			flags.do_minute_work = 0;
			minute_work();
		}

		/* hour work */
		if (flags.do_hour_work) {
			flags.do_hour_work = 0;
			hour_work();
		}
	}
}



void set_mode(char x)
{
	mode = x;

	while (1) {

		switch (mode) {

		case MODE_OFF:
		case MODE_AUTO_OFF:
			dcf77_enable(0);
			tubes_enable(0);
			return;

		case MODE_TIME:
			dcf77_enable(0);
			time_refresh();
			tubes_enable(1);
			return;

		case MODE_DCF:
		case MODE_AUTO_DCF:
			tubes_enable(0);
			dcf77_enable(1);
			return;

		case MODE_EDITOR:
			dcf77_enable(0);
			editor_get();
			editor_refresh();
			tubes_enable(1);
			return;

		case MODE_PREVENT:
			dcf77_enable(0);
			tubes_enable(1);
			tubes_prevention_enable();
			return;

		case MODE_AUTO:
			switch (tube_save_check()) {

			case 0:
				mode = MODE_AUTO_OFF;
				break;
			case 1:
				mode = MODE_TIME;
				break;
			case 2:
				mode = MODE_PREVENT;
				break;
			}
			break;
		}
	}
}
