/*
 * $Id: time.c,v 1.21 2008/05/10 19:07:09 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18cxxx.h>
#include "NixieClock.h"
#include "time.h"
#include "ds1307.h"
#include "tubes.h"
#include "util.h"
#include "editor.h"



/* global variables */
volatile time_t time;
volatile date_t date;
volatile tube_time_t tube_time;
volatile unsigned char tick;			/* 1/256s */
volatile unsigned char beep_duration;
hm_t alarm;
hm_t tube_off;
hm_t tube_on;
hm_t t_now;


/* prototypes */
static void increment_tube_time(void);
static char compare_hm(hm_t *, hm_t *);


/*
 * Set current time
 */
signed char set_time(volatile time_t *now)
{
	/* software time */
	time.hour = now->hour;
	time.min = now->min;
	time.sec = now->sec;

	/* DS1307 */
	return (ds1307_set_time(now));
}


/*
 * Set current date
 */
signed char set_date(volatile date_t *today)
{
	/* software date */
	date.year = today->year;
	date.month = today->month;
	date.day = today->day;
	date.weekday = today->weekday;

	/* DS1307 */
	return (ds1307_set_date(today));
}


/*
 * This routine is called from timer1 when 1 second has
 * elapsed.
 */
void time_inc(void)
{
	static char ret;

	/* keep track of current time */
	ret = increment_time(&time);
	if (ret & MIN_OVERLAP) {
		/* flag work to do in main loop (outside interrupt) */
		flags.do_minute_work = 1;
	}
	if (ret & HOUR_OVERLAP) {
		/* flag work to do in main loop (outside interrupt) */
		flags.do_hour_work = 1;
	}

	/* display time in TIME mode */
	if (mode == MODE_TIME) {
		time_refresh();
	}

	/* editor timeout (in seconds) */
	if (mode == MODE_EDITOR) {
		if (ed_timeout == 0) {
			/* exit EDITOR mode after timeout */
			editor_timeout();
		} else {
			ed_timeout--;
		}
	}

	/* cathode poisoning prevention */
	if (mode == MODE_PREVENT) {
		tubes_prevention_refresh();
	}

	/* increment tube time */
	increment_tube_time();
}


/* increment time by 1 second. Return 1 if the day changes */
signed char increment_time(volatile time_t *x)
{
	char ret = 0;

	x->sec++;
	if (x->sec == 60) {
		x->sec = 0;
		ret |= MIN_OVERLAP;
		x->min++;
		if (x->min == 60) {
			x->min = 0;
			ret |= HOUR_OVERLAP;
			x->hour++;
			if (x->hour == 24) {
				x->hour = 0;
				ret |= DAY_OVERLAP;
			}
		}
	}

	return (ret);
}


/* increment tube time by 1 sec */
static void increment_tube_time(void)
{
	if (HV_PIN == 0) {
		/* tubes are OFF. */
		return;
	}

	tube_time.sec++;
	if (tube_time.sec == 60) {
		tube_time.sec = 0;
		/* set flag to save the value to NVRAM */
		flags.do_save_tube_time = 1;
		tube_time.min++;
		if (tube_time.min == 60) {
			tube_time.min = 0;
			tube_time.hour++;
			if (tube_time.hour == 24) {
				tube_time.hour = 0;
				tube_time.day++;
			}
		}
	}
}


/*
 * show_time : set tubes_raw based on current time
 */
void time_refresh(void)
{
	static volatile time_t time1;
	static time_t time0;

	/* snapshot time */
	PIE1bits.TMR1IE = 0;
	time0 = time;
	PIE1bits.TMR1IE = 1;

	/* time + 1 sec for fading */
	time1 = time0;
	increment_time(&time1);

#if USE_DP
	tubes_dp = 0;
#endif

	/* show date or brightness temporary ? */
	if (temp_show > 0) {
		if (flags.temp_brightness) {
			/* HH MM _B */
			tubes_set_pair(T_1, 0, time0.hour, prefs.hour_leading_zero);
			tubes_set_pair(T_1, 1, time0.min, 0);
			tubes_set_pair(T_2, 0, time1.hour, prefs.hour_leading_zero);
			tubes_set_pair(T_2, 1, time1.min, 0);
			tubes_set_pair(T_BOTH, 2, brightness, 1);
			return;
		}

		if (flags.temp_date) {
			if (prefs.date_format) {
				/* MM DD YY */
				tubes_set_pair(T_BOTH, 1, date.day, 0);
				tubes_set_pair(T_BOTH, 0, date.month, 0);
			} else {
				/* DD MM YY */
				tubes_set_pair(T_BOTH, 0, date.day, 0);
				tubes_set_pair(T_BOTH, 1, date.month, 0);
			}
			tubes_set_pair(T_BOTH, 2, date.year, 0);

			return;
		}
	} else {
		flags.temp_brightness = 0;
		flags.temp_date = 0;
	}

	/* do we need to show the date ? */
	if (show_date_duration != 0) {

		static unsigned char end;
		static unsigned char day_pos;
		static unsigned char month_pos;

		day_pos = prefs.date_format ? 1 : 0;
		month_pos = prefs.date_format ? 0 : 1;

		end = show_date_when + show_date_duration;
		if (end > 59) {
			end -= 59;
		}

		if (time0.sec == show_date_when) {
			tubes_set_pair(T_1, 0, time0.hour, prefs.hour_leading_zero);
			tubes_set_pair(T_1, 1, time0.min, 0);
			tubes_set_pair(T_1, 2, time0.sec, 0);
			tubes_set_pair(T_2, day_pos, date.day, 0);
			tubes_set_pair(T_2, month_pos, date.month, 0);
			tubes_set_pair(T_2, 2, date.year, 0);
			return;
		}

		if (((end > show_date_when) &&
			(time0.sec > show_date_when && time0.sec < end)) ||
		    ((end < show_date_when) &&
			(time0.sec > show_date_when || time0.sec < end))) {

			tubes_set_pair(T_BOTH, day_pos, date.day, 0);
			tubes_set_pair(T_BOTH, month_pos, date.month, 0);
			tubes_set_pair(T_BOTH, 2, date.year, 0);
			return;
		}

		if (time0.sec == end) {
			tubes_set_pair(T_1, day_pos, date.day, 0);
			tubes_set_pair(T_1, month_pos, date.month, 0);
			tubes_set_pair(T_1, 2, date.year, 0);
			tubes_set_pair(T_2, 0, time1.hour, prefs.hour_leading_zero);
			tubes_set_pair(T_2, 1, time1.min, 0);
			tubes_set_pair(T_2, 2, time1.sec, 0);
			return;
		}
	 }

	/* default : show time */
	tubes_set_pair(T_1, 0, time0.hour, prefs.hour_leading_zero);
	tubes_set_pair(T_1, 1, time0.min, 0);
	tubes_set_pair(T_1, 2, time0.sec, 0);
	tubes_set_pair(T_2, 0, time1.hour, prefs.hour_leading_zero);
	tubes_set_pair(T_2, 1, time1.min, 0);
	tubes_set_pair(T_2, 2, time1.sec, 0);
}



/* trigger alarm */
void start_alarm(void)
{
	flags.do_alarm = 1;
	beep_duration = alarm_duration;
}

void stop_alarm(void)
{
	flags.do_alarm = 0;
	beep_duration = 0;
	ALARM_PIN = 0;
}


/*
 * This routine is called every minute.
 *
 * Check if tubes should be turned off, alarm, etc...
 */
void minute_work(void)
{
	static char r;

	/* check for alarm (all modes) */
	if (prefs.alarm_enable) {
		if (time.hour == alarm.hour && time.min == alarm.min) {
			start_alarm();
		}
	}

	/* refresh date */
	ds1307_get_date(&date);


	/* check if we need to go to some other mode */
	r = tube_save_check();

	/*
	 * in all mode except OFF, do cathode poisoning prevention if
	 * selected
	 */
	if (mode != MODE_OFF) {
		if (t_time_reload != 0 && r == 2) {
			set_mode(MODE_PREVENT);
			return;
		}
	}

	/* in TIME mode, check if the tubes should be turned OFF */
	if (mode == MODE_TIME) {
		if (r == 0) {
			if (prefs.tube_save_dcf) {
				set_mode(MODE_AUTO_DCF);
			} else {
				set_mode(MODE_AUTO_OFF);
			}
			return;
		}
	}

	/*
	 * - in AUTO_OFF and AUTO_DCF check if the tubes should be turned ON
	 * - in OFF mode, the tubes remain OFF
	 * - in DCF mode (manual), tubes should also remain OFF.
	 */
	if (mode == MODE_AUTO_OFF || mode == MODE_AUTO_DCF) {
		if (r == 1) {
			set_mode(MODE_TIME);
			return;
		}
	}
}



/*
 * Check if the tubes should be turned off or on automatically
 *
 * Returns :
 *   0   tubes should be off
 *   1   tubes should be on
 *   2   tubes cathode poisoning should start now
 */
char tube_save_check(void)
{
	static char c1, c2;

	if (t_time_reload != 0) {
		if (time.hour == t_when.hour && time.min == t_when.min) {
			return (2);
		}
	}

	/* if power saving feature is off, tubes should be always ON */
	if (!prefs.tube_save) {
		return (1);
	}

	t_now.hour = time.hour;
	t_now.min = time.min;

	c1 = (compare_hm(&t_now, &tube_off) < 0);
	c2 = (compare_hm(&t_now, &tube_on) >= 0);

	if (compare_hm(&tube_off, &tube_on) > 0) {
		/* example  22:00 -- 04:55 */
		return (c1 && c2);
	} else {
		/* example   01:00 -- 07:30 */
		return (c1 || c2);
	}
}


/*
 * This routine is called every hour
 */
void hour_work(void)
{

}




/*
 * Compare 2 hm_t structures.
 *
 * Returns  1 if t1 >  t2
 *         -1 if t1 <  t2
 *          0 if t1 == t2
 */
static char compare_hm(hm_t *t1, hm_t *t2)
{
	if (t1->hour > t2->hour) {
		return (1);
	}
	if (t1->hour < t2->hour) {
		return (-1);
	}
	if (t1->min > t2->min) {
		return (1);
	}
	if (t1->min < t2->min) {
		return (-1);
	}
	return (0);
}
