/*
 * $Id: tubes.c,v 1.11 2008/06/11 15:42:42 fvecoven Exp $
 *
 * Copyright (C) 2008 Frederic Vecoven
 *
 * This file is part of NixieClock
 *
 * NixieClock is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NixieClock is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <p18cxxx.h>
#include <delays.h>
#include "NixieClock.h"
#include "tubes.h"
#include "time.h"
#include "util.h"


/* global variables */
volatile unsigned char tubes_raw[3];
volatile unsigned char tubes_2_raw[3];
volatile unsigned char refresh_count;
volatile unsigned char current_tubes;
volatile unsigned char brightness;
volatile unsigned char brightness_refresh;
volatile unsigned char fade_threshold;
volatile unsigned char t_prevent;
volatile unsigned int t_time;
volatile unsigned char t_time_reload;
volatile unsigned char t_current;
volatile hm_t t_when;

#if USE_DP
volatile unsigned char tubes_dp;
#endif


/*
 * Hardware setup :
 *
 * Convention : tube 0 is the left one, tube 5 is the rightmost one.
 *
 * Assuming tubes_raw[3] = { 0x30, 0x41, 0x52 }, you will see
 * 0 1 2 3 4 5 on the tubes.
 *
 * Decimal point DP1 controls tubes 0 1 2
 * Decimal point DP2 controls tubes 3 4 5
 *
 * tubes_dp b'xx543210' where 0 controls tube 0, etc...
 */


/*
 * tubes ON/OFF
 */
void tubes_enable(char enable)
{
	if (enable) {
		DP1 = 0;
		DP2 = 0;
		HV_ON;
		/*
		 * delay to allow HV capacitor to charge before
		 * turning the tubes on.
		 *
		 * At 32MHz, 200ms = 16E6 cycles
		 */
		Delay10KTCYx(160);
		flags.do_tube = 1;
	} else {
		flags.do_tube = 0;
		ANODE0 = 0;
		ANODE1 = 0;
		ANODE2 = 0;
		DP1 = 1;
		DP2 = 1;
		HV_OFF;
	}
}


/* refresh routine
 *
 * - called every 100us from timer0 interrupt
 * - one tube gets 32 x 100us + 100us blanking = 3.3ms
 *   giving a refresh rate of 101Hz
 */
void tubes_refresh(void)
{

	/* check if we need to handle the next tube */
	if (refresh_count == 0) {
		if (current_tubes == 0) {
			ANODE1 = 0;
			ANODE2 = 0;
			ANODE0 = 1;
		} else if (current_tubes == 1) {
			ANODE0 = 0;
			ANODE2 = 0;
			ANODE1 = 1;
		} else {
			ANODE0 = 0;
			ANODE1 = 0;
			ANODE2 = 1;
		}
	}

	/* blanking and brightness */
	if (refresh_count > brightness_refresh) {
		ANODE0 = 0;
		ANODE1 = 0;
		ANODE2 = 0;
		PORTB = 0xFF;
		DP1 = 0;
		DP2 = 0;
	} else {

		/*
		 * fading : 1/2s value1, then value2 replaces value1
		 * progressively.
		 * The threshold is computed in timer1 because it
		 * varies based on tick
		 */
		if (refresh_count >= fade_threshold) {
			PORTB = tubes_raw[current_tubes];
		} else {
			PORTB = tubes_2_raw[current_tubes];
		}

#if USE_DP
		/* decimal points */
		if (tubes_dp & (1 << current_tubes)) {
			DP1 = 1;
		} else {
			DP1 = 0;
		}
		if ((tubes_dp >> 4) & (1 << current_tubes)) {
			DP2 = 1;
		} else {
			DP2 = 0;
		}
#endif
	}

	/* prepare next refresh */
	refresh_count++;
	if (refresh_count == 33) {
		refresh_count = 0;
		current_tubes++;
		if (current_tubes == 3) {
			current_tubes = 0;
		}
	}
}


/*
 * set brightness : 0 -> 7 (so brightness_refresh is 0 -> 31)
 *                  8 : auto-brightness
 */
void tubes_brightness(unsigned char value)
{
	if (value > 7) {
		prefs.auto_brightness = 1;
		flags.do_ad = 1;
	} else {
		prefs.auto_brightness = 0;
		brightness = value;
		brightness_refresh = (value + 1) << 2;
	}
}


/* adjust brightness. +1 if x > 0, -1 otherwise. */
void tubes_adjust_brightness(signed char x)
{
	if (x > 0) {
		if (brightness != 8) brightness++;
	} else {
		if (brightness != 0) brightness--;
	}
	tubes_brightness(brightness);
	flags.temp_brightness = 1;
	temp_show = 128;
	time_refresh();
}


/*
 * Utility function to set the value of a pair of tubes.
 *
 * Tubes 01 23 45   (pairs 0 1 2)
 *
 * Set the value of one tube-pair (01, 23 or 45) based on a
 * integer (0->99). If 'zero' is set, leading 0 is omitted.
 *
 * sel : 1 = tubes_raw  2 = tubes_2_raw  3 = both
 */
void tubes_set_pair(char sel, char which, unsigned char x,
	char zero)
{
	static unsigned char *ptr;
	static char i;

	if (x > 99) {
		x = 0xFF;
	} else {
		x = bcd(x);
		if (zero && !(x & 0xF0)) {
			x |= 0xF0;
		}
	}

	for (i = 0; i < 2; i++) {

		if (sel & (1 << i)) {
			ptr = (i == 0) ? tubes_raw : tubes_2_raw;

			switch (which) {

			case 0:		// left-most pair
				ptr[0] &= 0xF0;
				ptr[0] |= (x >> 4);
				ptr[1] &= 0xF0;
				ptr[1] |= (x & 0x0F);
				break;

			case 1:		// middle pair
				ptr[0] &= 0x0F;
				ptr[0] |= (x << 4);
				ptr[2] &= 0xF0;
				ptr[2] |= (x >> 4);
				break;

			case 2:		// right-most pair
				ptr[1] &= 0x0F;
				ptr[1] |= (x & 0xF0);
				ptr[2] &= 0x0F;
				ptr[2] |= (x << 4);
				break;
			}
		}
	}

}


/* utility function to set the value of one tube (numbered 0 1 2 3 4 5) */
void tubes_set_one(char sel, char which, unsigned char x)
{
	static unsigned char *ptr;
	static char i;

	if (x > 9) {
		x = 0x0F;
	}

	for (i = 0; i < 2; i++) {

		if (sel & (1 << i)) {
			ptr = (i == 0) ? tubes_raw : tubes_2_raw;

			switch (which) {

			case 0:		// left-most tube
			case 1:
			case 2:
				ptr[which] &= 0xF0;
				ptr[which] |= x;
				break;
			case 3:
			case 4:
			case 5:
				ptr[which-3] &= 0x0F;
				ptr[which-3] |= (x << 4);
				break;
			}
		}
	}
}



/*
 * Cathode poisoning prevention routine
 *
 * - do one pair of tubes at a time (to maximize current)
 * - cycle numbers 9 -> 0 during t_time_reload (user setting)
 */
void tubes_prevention_enable(void)
{
	t_prevent = 9;
	t_time = 10 * t_time_reload;
	t_current = 3;
	flags.do_tube = 0;	// stop tube multiplex
	tubes_prevention_refresh();
}


void tubes_prevention_refresh(void)
{
	if (t_current == 3) {
		ANODE0 = 0;
		ANODE1 = 0;
		ANODE2 = 1;
	} else if (t_current == 2) {
		ANODE0 = 0;
		ANODE2 = 0;
		ANODE1 = 1;
	} else {
		ANODE1 = 0;
		ANODE2 = 0;
		ANODE0 = 1;
	}

	PORTB = (t_prevent << 4) | t_prevent;

	t_time--;
	if (t_time == 0) {
		t_time = 10 * t_time_reload;
		t_prevent--;
		if (t_prevent == 0xFF) {
			t_prevent = 9;
			t_current--;
			if (t_current == 0) {
				/* we are done ! */
				set_mode(MODE_AUTO);
				return;
			}
		}
	}
}

